/*
 *  Error handling for libcdecl.
 *  Copyright © 2011-2012, 2021, 2023 Nick Bowler
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>

#include "cdecl.h"
#include "cdecl-internal.h"

#include <glthread/lock.h>
#include <glthread/tls.h>

#include "errmsg.h"

gl_once_define(static, tls_initialized)
static gl_tls_key_t tls_key;

struct err_state {
	struct cdecl_error err;
	size_t nstr;
	char str[FLEXIBLE_ARRAY_MEMBER];
};

/* This pre-initialized error is reserved for dire out-of-memory conditions. */
static struct cdecl_error err_no_mem;

static void free_err(void *err)
{
	if (err == &err_no_mem)
		return;

	free(err);
}

static void set_err(unsigned code, struct cdecl_error *err)
{
	static const char errmsgs[] = STRTAB_INITIALIZER;

	switch (code) {
	case CDECL__ENOMEM:
		err->code = CDECL_ENOMEM;
		break;
	default:
		err->code = CDECL_ENOPARSE;
		break;
	}

	err->str = _(&errmsgs[code]);
}

static void initialize(void)
{
	cdecl__init_i18n();
	set_err(CDECL__ENOMEM, &err_no_mem);
	gl_tls_key_init(tls_key, free_err);
}

static void *alloc_err_state(void *old, size_t buf_size)
{
	struct err_state *state;
	void *p;

	state = p = realloc(old, offsetof(struct err_state, str) + buf_size);
	if (state) {
		state->nstr = buf_size;
	} else if (old) {
		/* Failed allocation, but existing state is still good */
		p = old;
	} else {
		/* Failed allocation, no existing state */
		p = &err_no_mem;
	}

	gl_tls_set(tls_key, p);
	return state;
}

static struct err_state *get_err_state(void)
{
	void *state;

	gl_once(tls_initialized, initialize);

	state = gl_tls_get(tls_key);
	if (state == &err_no_mem)
		state = NULL;
	if (!state)
		return alloc_err_state(state, 100);
	return state;
}

/*
 * Set the library error to one of the preset messages defined in errmsg.h
 * (CDECL__Exxx).
 */
void cdecl__errmsg(unsigned msg)
{
	struct err_state *state;

	state = get_err_state();
	if (!state)
		return;

	set_err(msg, &state->err);
}

/*
 * Sets the library error to code, with a printf-style error string.
 */
void cdecl__err(unsigned code, const char *fmt, ...)
{
	struct err_state *state;
	int rc, try = 0;
	va_list ap;

	state = get_err_state();
	if (!state)
		return;

retry:
	va_start(ap, fmt);
	rc = vsnprintf(state->str, state->nstr, fmt, ap);
	va_end(ap);

	if (rc > 0 && rc >= state->nstr) {
		assert(try++ == 0 && rc < SIZE_MAX / 4);
		state = alloc_err_state(state, (size_t)(rc+1u) * 3 / 2);
		if (!state)
			return;

		goto retry;
	}

	state->err.str = state->str;
	state->err.code = code;
}

const struct cdecl_error *cdecl_get_error(void)
{
	struct err_state *state = get_err_state();

	return state ? &state->err : NULL;
}
